# -*- coding: utf-8 -*-
"""Consts for Dio Chacon wifi python client API."""
from enum import Enum

DIOCHACON_WS_URL = "wss://l4hfront-prod.chacon.cloud/ws"


class DeviceTypeEnum(Enum):

    def from_dio_api(label: str):
        dict_array = [
            {'label': ".wifi.shutter.", 'code': "SHUTTER"},
            {'label': ".wifi.genericSwitch.", 'code': "SWITCH_LIGHT"},
            {'label': ".wifi.plug.", 'code': "SWITCH_PLUG"},
        ]
        for d in dict_array:
            if d['label'] in label:
                return DeviceTypeEnum(d['code'])
        return DeviceTypeEnum.UNKNOWN

    def equals(self, other_label: str) -> bool:
        if DeviceTypeEnum.from_dio_api(other_label) == self:
            return True
        if other_label in [e.value for e in DeviceTypeEnum]:
            return True
        return False

    SHUTTER = "SHUTTER"
    SWITCH_LIGHT = "SWITCH_LIGHT"
    SWITCH_PLUG = "SWITCH_PLUG"
    UNKNOWN = "UNKNOWN"


class ShutterMoveEnum(Enum):
    UP = "up"
    STOP = "stop"
    DOWN = "down"


class SwitchOnOffEnum(Enum):
    ON = 1
    OFF = 0



"""Exceptions for DIOChacon API."""


class DIOChaconAPIError(Exception):
    """Error from this api."""

    def __init__(self, *args) -> None:
        Exception.__init__(self, *args)


class DIOChaconInvalidAuthError(Exception):
    """Invalid auth detected"""

    def __init__(self, *args) -> None:
        Exception.__init__(self, *args)



# coding: utf-8
"""Tests exceptions."""
import pytest
from dio_chacon_wifi_api.exceptions import DIOChaconAPIError
from dio_chacon_wifi_api.exceptions import DIOChaconInvalidAuthError


def test_exceptions() -> None:

    with pytest.raises(DIOChaconAPIError):
        raise DIOChaconAPIError("Dumb test coverage")

    with pytest.raises(DIOChaconInvalidAuthError):
        raise DIOChaconInvalidAuthError("Dumb 2")



# coding: utf-8
"""Tests session.py. DIOChaconClientSession class."""
import asyncio
import logging
from typing import Any

import pytest
from aiohttp_fake_server_utils import MOCK_PORT
from aiohttp_fake_server_utils import run_fake_http_server
from dio_chacon_wifi_api.session import DIOChaconClientSession

_LOGGER = logging.getLogger(__name__)

USERNAME = 'toto@toto.com'
PASSWORD = 'DUMMY_PASS'
SERVICE_NAME = "test_client"


@pytest.mark.asyncio
async def test_session(aiohttp_server) -> None:
    """Test generic for session."""

    recording_queue: asyncio.Queue = asyncio.Queue()
    message_queue: asyncio.Queue = asyncio.Queue()
    await run_fake_http_server(aiohttp_server, recording_queue)

    _LOGGER.debug("Launching test session...")

    def message_callback(data: Any) -> None:
        _LOGGER.info("******** CALLBACK MESSAGE RECEIVED *******")
        _LOGGER.info(data)
        _LOGGER.info("******** CALLBACK MESSAGE DONE *******")
        message_queue.put_nowait(data)

    session = DIOChaconClientSession(USERNAME, PASSWORD, SERVICE_NAME, message_callback)
    session._set_server_urls(f"ws://localhost:{MOCK_PORT}/ws")

    await session.ws_connect()

    message = await asyncio.wait_for(message_queue.get(), 2)
    assert message["name"] == 'connection'
    assert message["action"] == 'success'
    message_queue.task_done()

    await session.disconnect()



# -*- coding: utf-8 -*-
"""Client for the DIO Chacon wifi API."""

import asyncio
import logging
from asyncio import Lock
from asyncio import Queue
from typing import Any

from .const import DeviceTypeEnum
from .const import DIOCHACON_WS_URL
from .const import ShutterMoveEnum
from .const import SwitchOnOffEnum
from .exceptions import DIOChaconAPIError
from .exceptions import DIOChaconInvalidAuthError
from .session import DIOChaconClientSession

_LOGGER = logging.getLogger(__name__)


class DIOChaconAPIClient:
    """Client to the DIO Chacon wifi API.
    It is mainly a proxy to the chacon's cloud server.
    It manages authentication, the protocol through the websocket connection and server side callback events.
    """

    def __init__(
        self,
        login_email: str,
        password: str,
        service_name: str = "python_generic",
        callback_device_state: callable = None,
    ) -> None:
        """Initialize the client API. Actually do nothing but storing informations.
        The effective authentication and connection are lazyly achieved.

        Parameters:
            login_email: string containing your email in DIO app
            password: string containing your password in DIO app
            service_name: arbitrary string identifying this client
            callback_device_state: the callback method that will be called for server side events
        """
        self._login_email: str = login_email
        self._password: str = password
        self._service_name: str = service_name
        self._callback_device_state: callable = callback_device_state
        self._callback_device_state_by_device: dict[str, callable] = {}
        self._session: DIOChaconClientSession | None = None
        # Unique message id for request / response correlation
        self._id: int = 0
        # Queue to await connection response from server
        self._messages_connection_queue: Queue = Queue()
        # Queue to await responses for queries
        self._messages_responses_queue: Queue = Queue()
        # In case of server responses concurrency, uses this buffer to find correct message id response.
        self._messages_buffer: dict = dict()
        # Lock to prevent initialisation of WS connection concurrently
        self._init_lock: Lock = Lock()
        self._ws_url: str = DIOCHACON_WS_URL

    def set_callback_device_state(self, callback_device_state: callable) -> None:
        """Register after the constructor the global callback method that will be called for server side events"""
        self._callback_device_state = callback_device_state

    def set_callback_device_state_by_device(self, target_id, callback_device_state: callable) -> None:
        """Register the per device callback method that will be called for server side events"""
        self._callback_device_state_by_device[target_id] = callback_device_state

    def _set_server_urls(self, ws_url: str) -> None:
        # Simple method to easily mock the server url.
        self._ws_url = ws_url

    async def _get_or_init_session(self) -> None:
        if self._session and self._session.is_disconnected():
            _LOGGER.warning("You have been disconnected. Automatic reconnection...")
            self._session = None
        if self._session is None:
            async with self._init_lock:
                if self._session is None:
                    _LOGGER.debug("Session creation via init_session")

                    session = DIOChaconClientSession(
                        self._login_email, self._password, self._service_name, self._message_received_callback
                    )
                    # Stores session to be able to call disconnect whatever happens next (ok or ko auth)
                    self._session = session
                    session._set_server_urls(self._ws_url)

                    await session.ws_connect()

                    # Wait for the reception of the connection success message from the server.
                    try:
                        connection_message = await asyncio.wait_for(self._messages_connection_queue.get(), 10)
                        self._messages_connection_queue.task_done()

                        if (
                            "name" in connection_message
                            and connection_message["name"] == "connection"
                            and connection_message["action"] == "invalid"
                        ):
                            _LOGGER.debug("Invalid auth response received : %s", connection_message)
                            raise DIOChaconInvalidAuthError("Invalid username/password.")
                        # Do nothing of the connection successful message.

                    except TimeoutError:
                        _LOGGER.error("Error connecting to the server !")
                        raise DIOChaconAPIError("No connection aknowledge message received from the server !")

                    _LOGGER.debug("End of session creation via init_session")

    def _message_received_callback(self, data: Any) -> None:
        """The callback called whenever a server side message is received.
        Parameters:
            data: the message which is a from json converted dict.
        """

        if "name" in data and data["name"] == "connection":
            # Sends the connection message (success or invalid) in the dedicated queue
            self._messages_connection_queue.put_nowait(data)
            return

        if "id" in data:
            # Sends the response message which has the same id has the request
            self._messages_responses_queue.put_nowait(data)
            return

        if "name" in data and data["name"] == "deviceState" and data["action"] == "update":
            # Sends the device state pushed from the server to the calling client
            # Sends only pertinent data :
            result = {}
            device_data = data["data"]
            result["id"] = device_data["di"]
            result["connected"] = device_data["rc"] == 1
            for link in device_data["links"]:
                if link["rt"] == "oic.r.openlevel":
                    result["openlevel"] = link["openLevel"]
                if link["rt"] == "oic.r.movement.linear":
                    result["movement"] = link["movement"]
                if link["rt"] == "oic.r.switch.binary":
                    result["is_on"] = link["value"] == SwitchOnOffEnum.ON.value

            sent = False

            if self._callback_device_state:
                _LOGGER.debug("Sending global callback event.")
                self._callback_device_state(result)
                sent = True

            if result["id"] in self._callback_device_state_by_device:
                _LOGGER.debug("Sending callback event for device %s", result["id"])
                self._callback_device_state_by_device[result["id"]](result)
                sent = True

            if sent:
                return

        _LOGGER.warning("Unknown message received and dropped / no callback registered for this message : %s", data)

    def _get_next_id(self) -> int:
        self._id = self._id + 1
        return self._id

    async def _consume_message_queue_and_cache_it(self) -> None:
        while not self._messages_responses_queue.empty():
            data = await self._messages_responses_queue.get()
            msg_id: int = int(data["id"])
            cached = self._messages_buffer.get(msg_id)
            if not cached:
                self._messages_buffer[msg_id] = data
            self._messages_responses_queue.task_done()

    async def _get_message_response_with_id(self, message_id: int) -> Any:
        await self._consume_message_queue_and_cache_it()
        cached = self._messages_buffer.get(message_id)
        # Try another after a small sleep
        if not cached:
            await asyncio.sleep(0.5)
            await self._consume_message_queue_and_cache_it()

        cached = self._messages_buffer.get(message_id)
        if not cached:
            _LOGGER.error("No response received for message id : %s", message_id)
            raise DIOChaconAPIError("Unexpected situation : no response with correct id !")

        # Remove value from cache has it is considered as consumed
        self._messages_buffer.pop(message_id)

        _LOGGER.debug("Messages buffer in memory cached size = %s", len(self._messages_buffer))

        return cached

    async def _send_ws_message(self, method: str, path: str, parameters: Any) -> Any:
        req_id = self._get_next_id()

        # Constructs the message that will be formated in JSON by the DIOChaconClientSession object
        msg = {}
        msg["method"] = method
        msg["path"] = path
        msg["parameters"] = parameters
        msg["id"] = req_id

        _LOGGER.debug("WS request to send = %s", msg)
        await self._get_or_init_session()
        await self._session.ws_send_message(msg)

        raw_results = await self._get_message_response_with_id(req_id)

        _LOGGER.debug("WS response with result : %s", raw_results)

        if raw_results["status"] != 200:
            raise DIOChaconAPIError(f"Error during API call : {raw_results}")

        return raw_results

    async def disconnect(self) -> None:
        """Disconnects for the cloud server and properly closes the connection.
        It must be called at the of API usage or before python program ending.
        """
        if self._session:
            # Close the web socket
            await self._session.disconnect()

    async def get_user_id(self) -> str:
        """Search for the user technical id based on its authentification elements.

        Returns:
            A string for the unique user id from the server.
        """

        raw_results = await self._send_ws_message("GET", "/user", {})

        return raw_results["data"]["id"]

    async def search_all_devices(
        self, device_type_to_search: list[DeviceTypeEnum] = None, with_state: bool = False
    ) -> dict:
        """Search all the known devices with their states : positions for shutters and on/off for switches

        Parameters:
            device_type_to_search: the device type to search for. None means to return all type (SHUTTERS and SWITCHES)
            with_state: True to return the detailed states like shutter position and switches on or off.

        Returns:
            A list of tuples composed of id, name, type, openlevel and movement for shutter, is_on for switches.
        """

        raw_results = await self._send_ws_message("GET", "/device", {})

        results = dict()
        ids = []
        for device in raw_results["data"]:
            result = {}
            id = device["id"]
            type = device["type"]
            device_type = DeviceTypeEnum.from_dio_api(type)
            if device_type is not DeviceTypeEnum.UNKNOWN and (
                not device_type_to_search or (device_type in device_type_to_search)
            ):
                ids.append(id)
                result["id"] = id
                result["name"] = device["name"]
                result["type"] = device_type.value  # Converts type to our constant definition
                result["model"] = device["modelName"] + "_" + device["softwareVersion"]
                results[id] = result

        if with_state:
            details = await self.get_status_details(ids, device_infos=results)
            for id in ids:
                if id not in details:
                    continue

                results[id]["connected"] = details[id]["connected"]
                if "openlevel" in details[id]:
                    results[id]["openlevel"] = details[id]["openlevel"]
                    results[id]["movement"] = details[id]["movement"]
                if "is_on" in details[id]:
                    results[id]["is_on"] = details[id]["is_on"]

        return results

    async def get_status_details(self, ids: list, notifyCallback: bool = False, device_infos: dict = None) -> dict:
        """Retrieves the status detailed of devices ids given.

        Parameters:
            ids: the device ids to search details for.
            notifyCallback: True to notify the callback function par device.
            device_infos: the devices infos (name and model) for requested ids. Used only to produce a log.

        Returns:
            A list of tuples composed of id, connected ; openlevel and movement for shutter, is_on for switch.
        """

        parameters = {"devices": ids}
        raw_results = await self._send_ws_message("POST", "/device/states", parameters)

        results = dict()
        for device_key in raw_results["data"]:
            device_data = raw_results["data"][device_key]
            result = {}
            result["id"] = device_key

            if device_data is None:
                # The server sends no data on the device but it exists (e.g with a very old firmware),
                # so we consider it as disconnected as the DIO App.
                name = device_infos[device_key]["name"] if device_infos else "Unknown"
                device_type = device_infos[device_key]["type"] if device_infos else "Unknown"
                model = device_infos[device_key]["model"] if device_infos else "Unknown"
                _LOGGER.warn(
                    "The device '%s' ('%s', '%s', '%s') is not fully recognized. "
                    "Probably because of a too old firmware.",
                    name,
                    model,
                    device_type,
                    device_key,
                )

                result["connected"] = False
                result["openlevel"] = 0
                result["movement"] = ShutterMoveEnum.STOP.value
                result["is_on"] = SwitchOnOffEnum.ON.value
            else:
                # Nominal case
                result["connected"] = device_data["rc"] == 1
                for link in device_data["links"]:
                    if link["rt"] == "oic.r.openlevel":
                        result["openlevel"] = link["openLevel"]
                    if link["rt"] == "oic.r.movement.linear":
                        result["movement"] = link["movement"]
                    if link["rt"] == "oic.r.switch.binary":
                        result["is_on"] = link["value"] == SwitchOnOffEnum.ON.value

            results[device_key] = result

            # Send the update via the callback by device.
            if notifyCallback and device_key in self._callback_device_state_by_device:
                _LOGGER.debug("Sending callback status details for device %s", device_key)
                self._callback_device_state_by_device[device_key](result)

        return results

    async def move_shutter_direction(self, shutter_id: str, direction: ShutterMoveEnum) -> None:
        """Moves the given shutter in the given direction.

        Parameters:
            shutter_id: the device id defining the chosen shutter.
            direction: up, down or stop movement.
        """

        parameters = {"movement": direction.value.lower()}
        await self._send_ws_message("POST", f"/device/{shutter_id}/action/mvtlinear", parameters)

    async def move_shutter_percentage(self, shutter_id: str, openlevel: int) -> None:
        """Moves the given shutter at a given position.

        Parameters:
            shutter_id: the device id defining the chosen shutter.
            openlevel: the open level percentage between 0 and 100.
        """
        parameters = {"openLevel": openlevel}
        await self._send_ws_message("POST", f"/device/{shutter_id}/action/openlevel", parameters)

    async def switch_switch(self, switch_id: str, set_on: bool) -> None:
        """Switches on or off the given switch.

        Parameters:
            switch_id: the device id defining the chosen switch.
            set_on: on or off as desired state.
        """
        val = SwitchOnOffEnum.ON.value if set_on else SwitchOnOffEnum.OFF.value
        parameters = {"value": val}
        await self._send_ws_message("POST", f"/device/{switch_id}/action/switch", parameters)



# coding: utf-8
"""Tests consts."""
from dio_chacon_wifi_api.const import DeviceTypeEnum


def test_enum() -> None:
    """Test DeviceTypeEnum."""
    assert DeviceTypeEnum.from_dio_api(".dio1.wifi.genericSwitch.switch.").value == "SWITCH_LIGHT"
    assert DeviceTypeEnum.SWITCH_LIGHT != DeviceTypeEnum.SHUTTER
    assert DeviceTypeEnum.SWITCH_LIGHT.equals(".dio1.wifi.genericSwitch.switch.")
    assert DeviceTypeEnum.SWITCH_LIGHT.equals("SWITCH_LIGHT")
    assert not DeviceTypeEnum.SWITCH_LIGHT.equals("toto")
    assert DeviceTypeEnum("SWITCH_PLUG") in [DeviceTypeEnum.SWITCH_LIGHT, DeviceTypeEnum.SWITCH_PLUG]
    assert "SWITCH_PLUG" not in [DeviceTypeEnum.SWITCH_LIGHT, DeviceTypeEnum.SWITCH_PLUG]
    # ValueError : assert DeviceTypeEnum("toto") not in [DeviceTypeEnum.SWITCH_LIGHT, DeviceTypeEnum.SWITCH_PLUG]



# -*- coding: utf-8 -*-
"""Http server mocking requests/responses for test_client.py.
Inspired from : https://gist.github.com/ex3cv/2bbbbaf748c604fd737e1bb8372336e7
"""
import asyncio
import json
import logging
from asyncio import Queue
from functools import partial

import aiohttp
from aiohttp import web

_LOGGER = logging.getLogger(__name__)

MOCK_PORT = 38080

INVALID_PASSWORD = "PASS_INVALID_AUTH"


async def endpoint(
    request: web.Request,
    recording_queue: Queue,
) -> web.Response:

    _LOGGER.debug("MOCK server. Request received = %s", request)
    for_assertions = {}
    for_assertions["protocol"] = "HTTP"
    for_assertions["method"] = request.method
    for_assertions["rel_url"] = str(request.rel_url)
    for_assertions["body"] = str(await request.text())

    # record request for future inspection
    await recording_queue.put(for_assertions)
    if str(request.rel_url).startswith("/api/session/login"):
        if INVALID_PASSWORD in (await request.text()):
            _LOGGER.debug("MOCK server. Response with fake invalid authentication")
            return web.json_response(body='{"status":400,"data":"Invalid username/password."}', status=200)
        else:
            _LOGGER.debug("MOCK server. Response with fake session token")
            return web.json_response(body='{"status":200,"data":{"sessionToken":"r:myfakesessionToken"}}', status=200)

    _LOGGER.debug("MOCK server. Response None")
    return None


async def websocket_messages_handler(ws: aiohttp.web_ws.WebSocketResponse, recording_queue: Queue):
    async for msg in ws:
        _LOGGER.debug("MOCK Server WS : received message : %s", msg)
        if msg.type == aiohttp.WSMsgType.TEXT:

            content = json.loads(msg.data)
            path = content["path"]
            id = content["id"]

            for_assertions = {}
            for_assertions["protocol"] = "WS"
            for_assertions["method"] = content["method"]
            for_assertions["parameters"] = content["parameters"]
            for_assertions["path"] = path
            for_assertions["id"] = id
            # record request for future inspection
            await recording_queue.put(for_assertions)

            if path == "/user":
                response = {}
                response["id"] = id
                response["status"] = 200
                response["data"] = {}
                response["data"]["id"] = "mocked-user-id"
                response["data"]["name"] = "mocked-user-name"
                _LOGGER.debug("MOCK Server WS : response /user to send. %s", response)
                await ws.send_str(json.dumps(response))

            if path == "/device":
                response = {}
                response["id"] = id
                response["status"] = 200
                response["data"] = list()
                # Add one shutter
                response["data"].append({})
                response["data"][0]["id"] = "L4HActuator_idmock1"
                response["data"][0]["name"] = "Shutter mock 1"
                response["data"][0]["type"] = ".dio1.wifi.shutter.mvt_linear."
                response["data"][0]["modelName"] = "CERSwd-3B"
                response["data"][0]["softwareVersion"] = "1.0.6"
                # Add one switch
                response["data"].append({})
                response["data"][1]["id"] = "L4HActuator_idmock2"
                response["data"][1]["name"] = "Shutter mock 2"
                response["data"][1]["type"] = ".dio1.wifi.genericSwitch.switch."
                response["data"][1]["modelName"] = "CERNwd-3B"
                response["data"][1]["softwareVersion"] = "1.0.4"
                # Add one unknown device
                response["data"].append({})
                response["data"][2]["id"] = "L4HActuator_idmock3"
                response["data"][2]["name"] = "Shutter mock 3 unknown"
                response["data"][2]["type"] = ".dio1.camera.unknown"
                response["data"][2]["modelName"] = "CERNwd-3B"
                response["data"][2]["softwareVersion"] = "X.0.X"

                _LOGGER.debug("MOCK Server WS : response /device to send. %s", response)
                await ws.send_str(json.dumps(response))

            if path == "/device/states":
                response = {}
                response["id"] = id
                response["status"] = 200
                response["data"] = {}
                response["data"]["L4HActuator_idmock1"] = {}
                response["data"]["L4HActuator_idmock1"]["rc"] = 1
                response["data"]["L4HActuator_idmock1"]["links"] = list()
                response["data"]["L4HActuator_idmock1"]["links"].append({})
                response["data"]["L4HActuator_idmock1"]["links"][0]["rt"] = "oic.r.openlevel"
                response["data"]["L4HActuator_idmock1"]["links"][0]["openLevel"] = 75
                response["data"]["L4HActuator_idmock1"]["links"].append({})
                response["data"]["L4HActuator_idmock1"]["links"][1]["rt"] = "oic.r.movement.linear"
                response["data"]["L4HActuator_idmock1"]["links"][1]["movement"] = "stop"
                response["data"]["L4HActuator_idmock2"] = {}
                response["data"]["L4HActuator_idmock2"]["rc"] = 1
                response["data"]["L4HActuator_idmock2"]["links"] = list()
                response["data"]["L4HActuator_idmock2"]["links"].append({})
                response["data"]["L4HActuator_idmock2"]["links"][0]["rt"] = "oic.r.switch.binary"
                response["data"]["L4HActuator_idmock2"]["links"][0]["value"] = 0

                _LOGGER.debug("MOCK Server WS : response /device/states to send. %s", response)
                await ws.send_str(json.dumps(response))

            if path == "/device/L4HActuator_idmock1/action/mvtlinear":
                response = {}
                response["id"] = id
                response["status"] = 200

                _LOGGER.debug(
                    "MOCK Server WS : response /device/L4HActuator_idmock1/action/mvtlinear to send. %s", response
                )
                await ws.send_str(json.dumps(response))

                # simulates asynchronous send then send a status message for callback server side event
                async def coro_callback():
                    await asyncio.sleep(0.200)
                    response = {}
                    response["name"] = "deviceState"
                    response["action"] = "update"
                    response["data"] = {}
                    response["data"]["di"] = "L4HActuator_idmock1"
                    response["data"]["rc"] = 1
                    response["data"]["links"] = list()
                    response["data"]["links"].append({})
                    response["data"]["links"][0]["rt"] = "oic.r.openlevel"
                    response["data"]["links"][0]["openLevel"] = 69
                    response["data"]["links"].append({})
                    response["data"]["links"][1]["rt"] = "oic.r.movement.linear"
                    response["data"]["links"][1]["movement"] = "down"
                    _LOGGER.debug("MOCK Server WS : callback status to send. %s", response)
                    await ws.send_str(json.dumps(response))

                asyncio.create_task(coro_callback())

            if path == "/device/L4HActuator_idmock1/action/openlevel":
                response = {}
                response["id"] = id
                response["status"] = 200

                _LOGGER.debug(
                    "MOCK Server WS : response /device/L4HActuator_idmock1/action/openlevel to send. %s", response
                )
                await ws.send_str(json.dumps(response))

                # simulates asynchronous send then send a status message for callback server side event
                async def coro_callback():
                    await asyncio.sleep(0.200)
                    response = {}
                    response["name"] = "deviceState"
                    response["action"] = "update"
                    response["data"] = {}
                    response["data"]["di"] = "L4HActuator_idmock1"
                    response["data"]["rc"] = 1
                    response["data"]["links"] = list()
                    response["data"]["links"].append({})
                    response["data"]["links"][0]["rt"] = "oic.r.openlevel"
                    response["data"]["links"][0]["openLevel"] = 69
                    response["data"]["links"].append({})
                    response["data"]["links"][1]["rt"] = "oic.r.movement.linear"
                    response["data"]["links"][1]["movement"] = "down"
                    _LOGGER.debug("MOCK Server WS : callback status to send. %s", response)
                    await ws.send_str(json.dumps(response))

                asyncio.create_task(coro_callback())

            if path == "/device/L4HActuator_idmock2/action/switch":
                response = {}
                response["id"] = id
                response["status"] = 200

                _LOGGER.debug(
                    "MOCK Server WS : response /device/L4HActuator_idmock2/action/switch to send. %s", response
                )
                await ws.send_str(json.dumps(response))

                # simulates asynchronous send then send a status message for callback server side event
                async def coro_callback():
                    await asyncio.sleep(0.200)
                    response = {}
                    response["name"] = "deviceState"
                    response["action"] = "update"
                    response["data"] = {}
                    response["data"]["di"] = "L4HActuator_idmock2"
                    response["data"]["rc"] = 1
                    response["data"]["links"] = list()
                    response["data"]["links"].append({})
                    response["data"]["links"][0]["rt"] = "oic.r.switch.binary"
                    response["data"]["links"][0]["value"] = 1
                    _LOGGER.debug("MOCK Server WS : callback status to send. %s", response)
                    await ws.send_str(json.dumps(response))

                asyncio.create_task(coro_callback())

            if path == "/session/logout":
                response = {}
                response["id"] = id
                response["status"] = 200
                response["name"] = "disconnection"
                response["action"] = "success"
                await ws.send_str(json.dumps(response))
                await ws.close()
                break

    _LOGGER.debug('MOCK Server WS : Websocket connection closed')


async def websocket_handler(request: web.Request, recording_queue: Queue):
    _LOGGER.debug('MOCK Server WS : Websocket connection starting')
    ws = web.WebSocketResponse()
    await ws.prepare(request)
    _LOGGER.debug('MOCK Server WS : Websocket connection ready')
    if request.query.get("password") == INVALID_PASSWORD:
        _LOGGER.debug('MOCK Server WS : Sending invalid auth for connection failure')
        await ws.send_str('{"name":"connection","action":"invalid","data":""}')
    else:
        _LOGGER.debug('MOCK Server WS : Sending connection success')
        await ws.send_str('{"name":"connection","action":"success","data":""}')
    await asyncio.create_task(websocket_messages_handler(ws, recording_queue))

    return ws


async def run_fake_http_server(aiohttp_server, recording_queue: Queue) -> None:

    _LOGGER.debug("Starting Mock server...")

    app = web.Application()

    call = partial(endpoint, recording_queue=recording_queue)
    call_ws = partial(websocket_handler, recording_queue=recording_queue)
    app.add_routes([web.route("*", "/api/{tail:.*}", call), web.route("*", "/ws", call_ws)])

    await aiohttp_server(app, port=MOCK_PORT)



# coding: utf-8
"""Integration tests for DIO Chacon wifi api."""
import asyncio
import logging
import os
from typing import Any

import pytest
from dio_chacon_wifi_api import DIOChaconAPIClient
from dio_chacon_wifi_api.const import DeviceTypeEnum
from dio_chacon_wifi_api.const import ShutterMoveEnum

# Enter correct real values here for the tests to complete successfully with real Server calls.

USERNAME = os.environ.get('DIO_USERNAME')
PASSWORD = os.environ.get('DIO_PASSWORD')
MY_SHUTTER_ID = os.environ.get('DIO_SHUTTER_ID_TEST')

logging.basicConfig(level=logging.DEBUG)

_LOGGER = logging.getLogger(__name__)

if os.environ.get('PYDBG', None):
    # This code will block execution until you run your Python: Remote Attach
    port = 5678
    host = 'localhost'
    print(f"Blocking for remote attach for vscode {host} {port}")
    import debugpy

    debugpy.listen((host, port))
    debugpy.wait_for_client()


@pytest.mark.skip("Not an automated test but an example of usage with real values.")
@pytest.mark.asyncio
async def test_integration_simple() -> None:
    """Connect then lists all the devices."""

    assert USERNAME is not None, "Please set env var before running integration tests !"

    def log_callback(data: Any) -> None:
        _LOGGER.info("******** CALLBACK MESSAGE RECEIVED *******")
        _LOGGER.info(data)
        _LOGGER.info("******** CALLBACK MESSAGE DONE *******")

    # Init client
    client = DIOChaconAPIClient(USERNAME, PASSWORD, callback_device_state=log_callback)

    user_id = await client.get_user_id()
    _LOGGER.info(f"User Id retrieved : {user_id}")

    list_devices = await client.search_all_devices(device_type_to_search=[DeviceTypeEnum.SHUTTER], with_state=False)
    _LOGGER.info(f"Devices found : {list_devices}")

    my_device = list_devices[MY_SHUTTER_ID]
    _LOGGER.info(f"My device found {MY_SHUTTER_ID} : {my_device}")
    assert my_device['name'] == 'Test'
    assert my_device['type'] == DeviceTypeEnum.SHUTTER.value

    # get shutter position
    list_pos = await client.get_status_details([MY_SHUTTER_ID])
    _LOGGER.info(f"Positions found : {list_pos}")

    _LOGGER.info("---------- Moving UP ---------")
    await client.move_shutter_direction(MY_SHUTTER_ID, ShutterMoveEnum.UP)

    await asyncio.sleep(10)

    list_pos = await client.get_status_details([MY_SHUTTER_ID])
    _LOGGER.info(f"Positions found after move UP: {list_pos}")

    _LOGGER.info("---------- Moving 75% ---------")
    await client.move_shutter_percentage(MY_SHUTTER_ID, 75)

    await asyncio.sleep(10)

    list_pos = await client.get_status_details([MY_SHUTTER_ID])
    _LOGGER.info(f"Positions found after move 75%: {list_pos}")

    _LOGGER.info("End of integration tests : deconnecting...")

    await client.disconnect()



# coding: utf-8
"""Tests client.py. DIOChaconAPIClient class."""
import asyncio
import logging
from typing import Any

import pytest
from aiohttp_fake_server_utils import MOCK_PORT
from aiohttp_fake_server_utils import run_fake_http_server
from dio_chacon_wifi_api.client import DIOChaconAPIClient
from dio_chacon_wifi_api.const import ShutterMoveEnum
from dio_chacon_wifi_api.exceptions import DIOChaconInvalidAuthError

_LOGGER = logging.getLogger(__name__)

USERNAME = 'toto@toto.com'
PASSWORD = 'DUMMY_PASS'
SERVICE_NAME = "test_client"

INVALID_PASSWORD = 'PASS_INVALID_AUTH'


def log_callback(data: Any) -> None:
    _LOGGER.info("******** CALLBACK MESSAGE RECEIVED *******")
    _LOGGER.info(data)
    assert data["id"]
    _LOGGER.info("******** CALLBACK MESSAGE DONE *******")


@pytest.mark.asyncio
async def test_client_invalid_auth(aiohttp_server) -> None:
    """Test invalid auth exception for client giving bad credential."""

    recording_queue: asyncio.Queue = asyncio.Queue()
    await run_fake_http_server(aiohttp_server, recording_queue)

    _LOGGER.debug("Launching test client...")

    with pytest.raises(DIOChaconInvalidAuthError) as excinfo:
        client = DIOChaconAPIClient(USERNAME, INVALID_PASSWORD, SERVICE_NAME)
        client.set_callback_device_state(log_callback)
        client._set_server_urls(f"ws://localhost:{MOCK_PORT}/ws")
        await client.get_user_id()

    assert str(excinfo.value) == "Invalid username/password."

    _LOGGER.debug("Invalid auth test OK. Disconnecting...")

    await client.disconnect()


@pytest.mark.asyncio
async def test_client(aiohttp_server) -> None:
    """Test generic for client."""

    recording_queue: asyncio.Queue = asyncio.Queue()
    await run_fake_http_server(aiohttp_server, recording_queue)

    _LOGGER.debug("Launching test client...")

    client = DIOChaconAPIClient(USERNAME, PASSWORD, SERVICE_NAME)
    client.set_callback_device_state(log_callback)
    client._set_server_urls(f"ws://localhost:{MOCK_PORT}/ws")

    # Test get_user_id
    effective_response = await client.get_user_id()
    _LOGGER.debug("Assertion queue size = %s", recording_queue.qsize())
    effective_request = await asyncio.wait_for(recording_queue.get(), 2)
    assert effective_request["protocol"] == "WS"
    assert effective_request["method"] == "GET"
    assert effective_request["path"] == "/user"
    assert effective_request["parameters"] == {}
    assert effective_request["id"] == 1
    assert effective_response == "mocked-user-id"
    recording_queue.task_done()

    # Test search_all_devices
    effective_response = await client.search_all_devices(with_state=True)
    effective_request = await asyncio.wait_for(recording_queue.get(), 2)
    _LOGGER.debug("Request : %s .Response : %s", effective_request, effective_response)
    assert effective_request["protocol"] == "WS"
    assert effective_request["method"] == "GET"
    assert effective_request["path"] == "/device"
    assert effective_request["parameters"] == {}
    assert effective_request["id"] == 2
    recording_queue.task_done()

    effective_request = await asyncio.wait_for(recording_queue.get(), 2)
    _LOGGER.debug("Intermediairay WS Request : %s", effective_request)
    assert effective_request["protocol"] == "WS"
    assert effective_request["method"] == "POST"
    assert effective_request["path"] == "/device/states"
    assert effective_request["parameters"] == {'devices': ['L4HActuator_idmock1', 'L4HActuator_idmock2']}
    assert effective_request["id"] == 3
    recording_queue.task_done()

    assert len(effective_response) == 2
    assert effective_response["L4HActuator_idmock1"]["id"] == "L4HActuator_idmock1"
    assert effective_response["L4HActuator_idmock1"]["name"] == "Shutter mock 1"
    assert effective_response["L4HActuator_idmock1"]["type"] == "SHUTTER"
    assert effective_response["L4HActuator_idmock1"]["model"] == "CERSwd-3B_1.0.6"
    assert effective_response["L4HActuator_idmock1"]["connected"]  # == True
    assert effective_response["L4HActuator_idmock1"]["openlevel"] == 75
    assert effective_response["L4HActuator_idmock1"]["movement"] == "stop"
    assert effective_response["L4HActuator_idmock2"]["id"] == "L4HActuator_idmock2"
    assert effective_response["L4HActuator_idmock2"]["name"] == "Shutter mock 2"
    assert effective_response["L4HActuator_idmock2"]["type"] == "SWITCH_LIGHT"
    assert effective_response["L4HActuator_idmock2"]["model"] == "CERNwd-3B_1.0.4"
    assert effective_response["L4HActuator_idmock2"]["connected"]  # == True
    assert not effective_response["L4HActuator_idmock2"]["is_on"]  # == False
    # Device 3 is unknown so not in the response since it is filtered by the client.

    # Test move_shutter_direction
    await client.move_shutter_direction(shutter_id="L4HActuator_idmock1", direction=ShutterMoveEnum.DOWN)
    effective_request = await asyncio.wait_for(recording_queue.get(), 2)
    _LOGGER.debug("Request : %s", effective_request)
    assert effective_request["protocol"] == "WS"
    assert effective_request["method"] == "POST"
    assert effective_request["path"] == "/device/L4HActuator_idmock1/action/mvtlinear"
    assert effective_request["parameters"] == {'movement': "down"}
    assert effective_request["id"] == 4
    recording_queue.task_done()
    # Returns is also done via a callback

    # Test move_shutter_percentage
    await client.move_shutter_percentage(shutter_id="L4HActuator_idmock1", openlevel=31)
    effective_request = await asyncio.wait_for(recording_queue.get(), 2)
    _LOGGER.debug("Request : %s", effective_request)
    assert effective_request["protocol"] == "WS"
    assert effective_request["method"] == "POST"
    assert effective_request["path"] == "/device/L4HActuator_idmock1/action/openlevel"
    assert effective_request["parameters"] == {'openLevel': 31}
    assert effective_request["id"] == 5
    recording_queue.task_done()
    # Returns is also done via a callback

    # Test switch_switch
    await client.switch_switch(switch_id="L4HActuator_idmock2", set_on=True)
    effective_request = await asyncio.wait_for(recording_queue.get(), 2)
    _LOGGER.debug("Request : %s", effective_request)
    assert effective_request["protocol"] == "WS"
    assert effective_request["method"] == "POST"
    assert effective_request["path"] == "/device/L4HActuator_idmock2/action/switch"
    assert effective_request["parameters"] == {'value': 1}
    assert effective_request["id"] == 6
    recording_queue.task_done()
    # Returns is also done via a callback

    # Explicitly awaits that the callbacks are executed by the fake server before disconnecting.
    _LOGGER.debug("Waiting for callbacks executions...")
    await asyncio.sleep(0.500)

    await client.disconnect()



# -*- coding: utf-8 -*-
"""Session manager for the DIO Chacon wifi API in order to maintain authentication token between calls.
And context between the various websockets calls.
Largely inspired by https://github.com/jjlawren/python-plexwebsocket/blob/master/plexwebsocket.py
"""
import asyncio
import json
import logging
import urllib

import aiohttp


MAX_FAILED_ATTEMPTS = 5

STATE_CONNECTED = "connected"
STATE_DISCONNECTED = "disconnected"
STATE_STARTING = "starting"
STATE_STOPPED = "stopped"

_LOGGER = logging.getLogger(__name__)


class DIOChaconClientSession:
    """HTTP session manager for DIO Chacon API.

    This session object allows to manage the authentication
    in the API using a token.
    """

    _state: str = None
    _failed_attempts: int = 0
    _aiohttp_session: aiohttp.ClientSession | None = None
    _websocket: aiohttp.ClientWebSocketResponse | None = None
    _listen_task: asyncio.Task | None = None

    def __init__(self, login_email: str, password: str, service_name: str, callback: callable) -> None:
        """Initialize and authenticate.

        Parameters:
            username: the dio chacon user (mainly the email to log in the app)
            password: the dio chacon user's password
            service_name: arbitrary string identifying the client
            callback (Runnable):
                Called when interesting events occur. Provides arguments:
                   data (str): websocket payload contents deserialized from json
        """
        self._login_email = login_email
        self._password = password
        self._service_name = service_name
        self._callback = callback

        async def on_request_start(session, trace_config_ctx, params):
            _LOGGER.debug(f"aiohttp request start : {params}")

        async def on_request_chunk_sent(session, trace_config_ctx, params):
            _LOGGER.debug(f"aiohttp request chunk sent : {params}")

        async def on_response_chunk_received(session, trace_config_ctx, params):
            _LOGGER.debug(f"aiohttp response chunk received : {params}")

        async def on_request_end(session, trace_config_ctx, params):
            _LOGGER.debug(f"aiohttp request end : {params}")

        trace_config = aiohttp.TraceConfig()
        trace_config.on_request_start.append(on_request_start)
        trace_config.on_request_end.append(on_request_end)
        trace_config.on_request_chunk_sent.append(on_request_chunk_sent)
        trace_config.on_response_chunk_received.append(on_response_chunk_received)
        self._aiohttp_session = aiohttp.ClientSession(trace_configs=[trace_config])

    def _set_server_urls(self, ws_url: str) -> None:
        # Simple method to easily mock the server url by overriding default values.
        self._ws_url = ws_url

    async def ws_connect(self) -> None:
        """Make a connection to the server via websocket protocol."""

        # Starts a background parallel task to permanently listen to events
        # sent by the WS server and transfer them via callback.
        self._listen_task = asyncio.create_task(self._listen())

    async def _listen(self) -> None:
        """Listen to messages"""
        # Infinite loop to listen to messages on the websocket and manage retries.
        self._failed_attempts = 0
        while self._state != STATE_STOPPED:
            await self._running()

    async def _running(self) -> None:
        url = (
            self._ws_url
            + "?"
            + urllib.parse.urlencode(
                {
                    'email': self._login_email,
                    'password': self._password,
                    'serviceName': self._service_name,
                }
            )
        )

        self._state = STATE_STARTING

        try:
            async with self._aiohttp_session.ws_connect(url, heartbeat=15, autoping=True) as ws_client:
                self._state = STATE_CONNECTED
                self._failed_attempts = 0
                self._websocket = ws_client
                async for message in ws_client:
                    if self._state == STATE_STOPPED:
                        break

                    if message.type == aiohttp.WSMsgType.ERROR:
                        _LOGGER.error("Error received from WS server: %s", message)
                        break

                    if message.type == aiohttp.WSMsgType.CLOSED:
                        _LOGGER.warning("AIOHTTP websocket connection closed")
                        break

                    if message.type == aiohttp.WSMsgType.TEXT:
                        _LOGGER.debug("Websocket received data %s", message)
                        msg = message.json()
                        self._callback(msg)

        except aiohttp.ClientResponseError as error:
            _LOGGER.error("Unexpected response received from server : %s", error)
            self._state = STATE_STOPPED
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError) as error:
            if self._failed_attempts >= MAX_FAILED_ATTEMPTS:
                _LOGGER.error("Too many retries to reconnect to server. Please restart globally.")
                self._state = STATE_STOPPED
            elif self._state != STATE_STOPPED:
                retry_delay = min(2 ** (self._failed_attempts - 1) * 30, 300)
                self._failed_attempts += 1
                _LOGGER.error("Websocket connection failed, retrying in %ds: %s", retry_delay, error)
                self._state = STATE_DISCONNECTED
                await asyncio.sleep(retry_delay)
        except Exception as error:
            if self._state != STATE_STOPPED:
                _LOGGER.exception("Unexpected exception occurred: %s", error)
                self._state = STATE_STOPPED

    async def disconnect(self) -> None:
        _LOGGER.debug("Disconnection of the current session")
        self._state = STATE_STOPPED
        if self._websocket:
            await self._websocket.close()
        if self._aiohttp_session:
            await self._aiohttp_session.close()
        # Let the _listen infinite loop terminate correctly with STATE_STOPPED signal.
        await asyncio.sleep(0.5)
        _LOGGER.debug("Disconnection done")

    async def ws_send_message(self, msg) -> None:
        """Sends a message in the websocket by converting it to json.

        Parameters:
            msg: the message to be sent
        """
        await self._websocket.send_str(json.dumps(msg))

    def is_disconnected(self) -> bool:
        return self._state == STATE_STOPPED or self._websocket.closed
